import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:open_weather_map_client/src/common/constants.dart';
import 'package:open_weather_map_client/src/common/parse_data.dart';
import 'package:open_weather_map_client/src/model/city.dart';
import 'package:open_weather_map_client/src/model/weather.dart';
import 'package:http/http.dart' as http;

/// Useful to obtain weather data, it connects with the Open Weather Map API.
class OpenWeatherMapProvider with ChangeNotifier {
  /// The key obtained from the Open Weather Map API.
  final String apiKey;

  /// Use the supported [Language] to get the data from the API.
  final Language language;

  /// If you don't need the five-day forecast, define it as false.
  ///
  /// Default is true.
  final bool loadForecast;

  // Controls the maximum time in seconds to wait to get the data.
  /// If you have a slow connection make sure to use the correct seconds.
  ///
  /// Default is 5 seconds
  final int requestTimeout;

  late City city;

  late Weather _weather;

  List<Weather>? _forecast;

  OpenWeatherMapProvider({
    required this.apiKey,
    this.language = Language.english,
    this.loadForecast = true,
    this.requestTimeout = 5,
  });

  Future<void> loadByCity({
    int? id,
    String? countryCode,
    String? languageCode,
    required String name,
  }) async {
    String body = await _response(
      Uri.https(
        apiAuthority,
        '$apiUnencodedPath$apiCurrentWeatherPath',
        _parameters(
          cityId: id,
          cityName: name,
          countryCode: countryCode,
          lang: languageCode,
        ),
      ),
    );
    _weather = await compute(parseWeather, body);
    if (loadForecast) {
      body = await _response(
        Uri.https(
          apiAuthority,
          '$apiUnencodedPath$apiFiveDayForecastPath',
          _parameters(
            cityId: id,
            cityName: name,
            countryCode: countryCode,
            lang: languageCode,
          ),
        ),
      );
      _forecast = await compute(parseForecast, body);
    }
  }

  Future<void> loadByLocation({
    required double latitude,
    required double longitude,
    String? languageCode,
  }) async {
    final Uri uri = Uri.https(
      apiAuthority,
      '$apiUnencodedPath$apiCurrentWeatherPath',
      _parameters(lang: languageCode, latitude: latitude, longitude: longitude),
    );
  }

  /// Parameters to use in URI for get request to API
  Map<String, String> _parameters({
    int? cityId,
    String? countryCode,
    String? cityName,
    double? latitude,
    double? longitude,
    String? lang,
  }) {
    String? city = cityName;
    if (city != null && countryCode != null) city += ',$countryCode';
    return {
      'appid': apiKey,
      if (cityId != null) 'id': cityId.toString(),
      if (city != null && cityId == null) 'q': city,
      if (city == null && cityId == null) 'lat': latitude.toString(),
      if (city == null && cityId == null) 'lon': longitude.toString(),
      'lang': lang ?? languageCode(language),
    };
  }

  /// Response request body from network as String
  Future<String> _response(Uri uri) async {
    return http
        .get(uri)
        .timeout(Duration(seconds: requestTimeout),
            onTimeout: () => throw TimeoutException(
                'The connection has timed out, Please try again!'))
        .then(
          (value) => value.body,
        );
  }

  /// Returns the current weather
  Weather get weather => _weather;

  /// Returns the five day forecast if [loadForecast] is true, otherwise it returns null.
  List<Weather>? get forecast => _forecast;
}
